# Event
#
# An Event changes the state of one or more simulation resources.
# Each event also has one or more processes.
#
# e.g. When modeling a queue, an Event would exist that denotes the
#      arrival of a new customer, thereby changing the state of the system
#      by incrementing the queue by 1.
#
# Instance Variables:
#
# - @name               Name of the event.
#                       e.g. @name = 'Customer Arrival'
#
# - @description        Description of the event.
#                       e.g. @description = 'The arrival of a single customer
#                                            to the queue.'
#
# - @processes          An array of Shrewd::Process objects that are to be
#                       executed following this event. These processes will
#                       schedule future events.
#                       e.g. @processes = [ Shrewd::Process.new('Process') ]
#
# - @action             A proc that is called to modify system and local state.
#                       This is the method by which events modify the system.
#                       The system state variables are injected.
#                       e.g. @action = Proc.new do |variables|
#                                        variables[:queue].decrement
#                                        variables[:employees].decrement
#                                      end

class Shrewd::Event
  attr_accessor :name, :description, :processes, :action

  # Initialize the event
  #
  # Input:
  # - name            name of the event
  # - description     description of the event (optional)
  # - processes       processes to be executed after event
  # - action          proc that changes the state of system variables
  def initialize(name, description = "", processes = [], action = nil)
    @name = name
    @description = description
    @processes = processes
    @action = action
  end

  # Trigger the event's action
  #
  # Input:
  # - state_variables       system state variables
  def trigger(state_variables)
    @action.call(state_variables) if @action
  end
end
