# Process
#
# A Process is triggered by an Event to schedule another Event.
# Each Process schedules an event in a given (or randomized) period of time.
# A Process can have conditions that determine if the associated Event
# should be queued.
#
# e.g. When modeling customer arrivals to a retail store, each customer
#      arrival Event would create a process that schedules a service Event,
#      in which a employee would service the customer.
#
# Instance Variables:
#
# - @name             Name of the process.
#                     e.g. @name = 'Service Customer'
#
# - @description      Description of the process.
#                     e.g. @description = 'An employee servicing a customer.'
#
# - @event            Event that the process triggers after checking conditions.
#                     e.g. @event = Shrewd::Event.new('Target')
#
# - @conditions       A proc that returns a boolean. If the result is true then
#                     the @event is added to the event queue. The proc is
#                     provided the system state variables.
#                     e.g. @conditions = Proc.new do |variables|
#                                                   if variables[:employees] > 0
#                                                     true
#                                                   else
#                                                     false
#                                                   end
#                                                 end
#
# - @delay            A proc that returns a number. This number determines how
#                     far in the future the event should be scheduled to
#                     execute (relative to current time).
#                     e.g. @delay = Proc.new do |variables|
#                                     if variables[:queue] > 5
#                                       15
#                                     else
#                                       10
#                                     end
#                                   end

class Shrewd::Process
  attr_accessor :name, :description, :event,
                :conditions, :delay

  # Initialize the Process
  #
  # Input:
  # - name                name of the process
  # - description         description of the process
  # - event               event that this process will trigger
  # - conditions          proc that returns bool to validate event scheduling
  # - delay               proc that returns the time to schedule the event
  def initialize(name, description = '', event = nil,
                 conditions = nil, delay = nil)
    @name = name
    @description = description
    @event = event
    @conditions = conditions
    @delay = delay
  end

  # Validate the conditions to determine if event should be scheduled
  def validated?(state_variables)
    if @conditions
      @conditions.call(state_variables)
    else
      true
    end
  end

  # Retrieve the calculated delay time
  def delay_time(state_variables)
    if @delay
      @delay.call(state_variables)
    else
      0
    end
  end
end
