# Simulation
#
# A Simulation is a container for a simulation that holds the current
# state of the system via state variables as well as holds the pending
# and historical events of the simulation.
#
# A simulation can be started and stopped with a Simulation instance.
#
# One or multiple initial Processes must be provided to the Simulation
# which will started at time 0.
#
# A Simulation can also be stepped through once stopped to observe the system
# state changes one event at a time.
#
# Instance Variables:
# - @variables          a hash of system variables whose key is the snake case
#                       name of the provided variable, and the value is the
#                       variable
#
#                       e.g. @variables = [ Shrewd::Variable.new('Queue')]
#
# - @initial_processes  the initial process
#                       e.g. @initial_processes = [Shrewd::Process.new('proc')]
#
# - @queue              array of hashes of events to be processed and their
#                       respective times of execution
#                       e.g. @queue = [ { time: @clock,
#                                         event: Shrewd::Event.new('Event') }]
#
# - @log                array of events and their time of execution
#                       e.g. @log << { event: Shrewd::Event.new, time: 15 }
#
# - @clock              int value representing current clock time
#                       e.g. @clock = 5
module Shrewd
  class Simulation
    attr_accessor :initial_processes

    # Initialize the simulation
    #
    # Input:
    # - variables     hash of the initial state of simulation state variables
    # - initial       initial process that will be triggered
    #
    # Return:
    # - [Shrewd::Simulation]: new simulation instance
    def initialize(variables = [], initial = nil)
      @variables          = {}
      @initial_processes  = initial
      @queue              = []
      @log                = []
      @clock              = 0

      # assign variables
      variables.each do |variable|
        name = variable.name.downcase.gsub(/\s+/,"_").to_sym
        @variables[name] = variable
      end

      # save initial state
      @initial_state = @variables
    end

    # Start the simulation with a default length of 1000
    def start(length = 1000)
      create_initial_event

      while @queue.any?
        # Set the clock
        event = @queue.shift
        @clock = event[:time].to_i
        break if @clock > length

        # Retrieve current event
        current_event = event[:event]
        current_event.trigger(@variables)

        # Queue events spawned by processes
        current_event.processes.each do |process|
          if process.validated?(@variables)
            process_time = process.delay_time(@variables)
            add_to_queue({ event: process.event, time: @clock + process_time })
          end
        end

        # Log the event
        @log << { event: current_event, time: @clock }
      end
    end

    # Print a log of the simulation events
    def print_log
      @log
    end

    # Reset the simulation
    def reset
      @log        = []
      @queue      = []
      @clock      = 0
      @variables  = @initial_state
    end

    private
      # Create the 'start' event of the simulation. The start event will call
      # the initial process.
      def create_initial_event
        name = 'Start'
        description = 'The first event of the simulation.'

        initial_event = Shrewd::Event.new(name, description)
        initial_event.processes = @initial_processes
        add_to_queue({ event: initial_event, time: @count })
      end

      # Add event to queue according to time of event.
      #
      # Input:
      # - event             the event/time hash to add to the queue
      def add_to_queue(event)
        size = @queue.size
        return @queue << event if size == 0

        # binary search
        index = search(event, 0, @queue.size - 1)
        if index > @queue.size - 1
          @queue.push(event)
        else
          @queue.insert(index, event)
        end
      end

      # Binary search to identify location in queue to add event, based on
      # the time the event will be triggered.
      #
      # Input:
      # - event           the event/time hash to add to the queue
      # - lo              low index in the search
      # - hi              high index in the search
      def search(event, lo, hi)
        # Low and high values
        low_time = @queue[lo][:time]
        high_time = @queue[hi][:time]

        # Event time is higher or lower than bounds
        return lo if low_time > event[:time]
        return hi + 1 if high_time <= event[:time]
        return lo if hi == lo

        # Midpoint
        midpoint = ((lo + hi) / 2).floor

        # Compare event to midpoint
        if event[:time] < @queue[midpoint][:time]
          search(event, 0, midpoint - 1)
        else
          search(event, midpoint + 1, @queue.size - 1)
        end
      end
  end
end

# require simulation entities
require_relative 'entities/event'
require_relative 'entities/process'
require_relative 'entities/variable'
